# This file contains some generic
# functions used by crtool.pl
#################################

# Extracts the Head of a mention
# (single token / compound proper noun)
#######################################
sub Head{

    my ($token, $tag) = @_;

    my $HeadNP = ();
    my $NP = ();
    
    # Proper Noun
    #------------

    # Person
    if ($tag =~ /^NP00SP0/) {
	# Compound Proper Noun: 1st token
	if ($token =~ /_/) {
	    (my @toks) = split("_", $token);
	    if ($toks[1] =~ /^d.s?$/) {
		$HeadNP = $toks[1]
	    } else {
		$HeadNP = $toks[0];
	    }
	}
	# Single Proper Noun, Head = token
	else {
	    $HeadNP = $token;
	}
    }
    return $HeadNP;
}


# Extracts Features
# Gender and Number
# (only gender now...)
###################
sub ExtractFeats {

    my ($pos, $Input, $token) = @_;
    my %Input = %$Input;
    my $gender = ();

    if (defined $FemNames{$token}) {
	$gender = "F";
    } elsif (defined $MaleNames{$token}) {
	$gender = "M";
    } else {
	if ($Input{$pos-1} =~ /^[0-9]/) {
	    if ($Input{$pos-1} =~ /(NCMS|AQ0MS|VMP00SM)/) {
		$gender = "M";
	    } elsif ($Input{$pos-1} =~ /(NCFS|AQ0FS|VMP00SF)/) {
		$gender = "F";
	    }
	}
    }
    return $gender;
}


# Looks for fixed structures of synonyms
########################################
sub Synonym {

    my ($Mentions, $Linhas, $m) = @_;

    my %Mentions = %$Mentions;
    my %Linhas = %$Linhas;
    my $used = 0;
    
    # Reverse position (for finding
    # the mention given the position
    my %PosMentions = reverse(%MentionsPos);

    # Line number of the proper noun
    my $n = $MentionsPos{$m} + 1;

    # Analyze next 5 lines
    my $end = $s + 5;

    for ($s = $n; $s <= $end; $s++) {
	if ($Linhas{$s} =~ /^[0-9]/) {
	    (@feats) = split(" ", $Linhas{$s});
	    my $pos = $feats[0];
	    my $tok = $feats[1];
	    my $lem = $feats[2];
	    my $tag = $feats[3];
	    
	    # Verbs of the structures ("X [, also] know as Y", etc.)
	    if ( ($tag =~ /^VMP/ && $lem =~ /^(conhecer|chamar|conocer|llamar|coñecer)$/) ||
		 ($tag =~ /^VBN/ && $lem =~ /^know$/) && ($s - $m < 3)
		) {
		# Max. distance of the second entity look-up
		my $e = $s + 4;
		
		# Not syntax look-up, just next tokens
		for (my $s2 = ($s + 1); $s2 <= $e; $s2++) {
		    my (@feats2) = split(" ", $Linhas{$s2});
		    my $tok2 = $feats2[1];
		    my $lem2 = $feats2[2];
		    my $tag2 = $feats2[3];

		    if ($tag2 =~ /^NP/) {
			$i = $PosMentions{$s2};
			$Mentions_tmp = Merge(\%Mentions, $m, $i);
			%Mentions_tmp = %$Mentions_tmp;
			$used = 1;
		    }
		}
	    }
	}
    }
    if ($used == 1) {
	if ($Mentions_tmp{$m} ne "X") {
	    #$TagEnt{$Mentions_tmp{$m}} = SelectTag($i, $m);
	    # Different order (i-m) for synonym structures
	    $TagEnt{$Mentions_tmp{$m}} = SelectTag($i, $m);
	}
	return \%Mentions_tmp;
    } else {
	return \%Mentions;
    }
}

# Compares gender (in the entity, if any) before the merging
############################################################
sub CompareGN {

    my ($m, $i) = @_;
    
    $GenOK = 0;
    my $Gen_m = ();
    my $Gen_i = ();

    if ($Mentions{$m} eq "X") {
	$Gen_m = $Gender{$m};
    } else {
	$Gen_m = $GenderEnt{$Mentions{$m}};
    }
    if ($Mentions{$i} eq "X") {
	$Gen_i = $Gender{$i};
    } else {
	$Gen_i = $GenderEnt{$Mentions{$i}};
    }

    # Only blocks different (M/F) genders
    if ( ($Gen_m eq "F" && $Gen_i eq "M") ||
	 ($Gen_m eq "M" && $Gen_i eq "F") ) {
	$GenOK = 0;
    } else {
	$GenOK = 1;
    }
    return $GenOK;
}

# Test: compare just from mentions, not entities
################################################
sub CompareNPMention {

    my ($m, $i) = @_;
    
    my $CompNP = 0;

    if ($MentionsForm{$m} =~ /(^|_|\-)$MentionsForm{$i}(_|\-|$)/) {
	$CompNP = 1;
    } elsif ($MentionsForm{$i} =~ /(^|_|\-)$MentionsForm{$m}(_|\-|$)/) {
	$CompNP = 1;
    }
    return $CompNP;
}

# Compares Proper nouns: Full Name and Tokens
#############################################
sub CompareNP {

    my ($m, $i) = @_;

    my $CompNP = 0;
    my $FullNP_m = ();
    my $FullNP_i = ();
    my $NP_m = ();
    my $NP_i = ();
    my $CompNP_toks = 0;
    my $ToksNP_m = 0;
    my $ToksNP_i = 0;
    my $LongNPsize_m = 0;
    my $LongNPsize_i = 0;
    my $LongNP_m = ();
    my $LongNP_i = ();

    # Chooses the longest proper noun of each mention
    # (then compared with the FullNP in the entity, if any)
    if ($FullNP{$m}) {
	$NP_m = $FullNP{$m};
	if ($NP_m =~ "_") {
	    my (@NPs_m) = split("_", $NP_m);
	    $ToksNP_m = $#NPs_m + 1;
	} else {
	    $ToksNP_m = 1;
	}
    } else {
	$ToksNP_m = 0;
	$NP_m = ();
    }
    if ($FullNP{$i}) {
	$NP_i = $FullNP{$i};
	if ($NP_i =~ "_") {
	    my (@NPs_i) = split("_", $NP_i);
	    $ToksNP_i = $#NPs_i + 1;
	} else {
	    $ToksNP_i = 1;
	}
    } else {
	$ToksNP_i = 0;
	$NP_i = ();
    }

    # Longest proper noun of the first entity
    if ($FullNPEnt{$Mentions{$m}}) {
	my %size = ();
	foreach my $f(@{$FullNPEnt{$Mentions{$m}}}) {
	    if ($f =~ /_/) {
		my (@size) = split("_", $f);
		$size{$f} = $#size + 1;
		my $first = 0;
		foreach my $n(sort {$size{$b} <=> $size{$a} } keys %size) {
		    if (!$first) {
			$LongNP_m = $n;
			$LongNPsize_m = $size{$n};
		    }
		    $first = 1;
		}
	    }
	}

	if ($LongNP_m) {
	    $FullNP_m = $LongNP_m;
	} else {
	    $LongNPsize_m = 1;
	}
    }

    # Entity to compare
    # (from entity or mention)
    my $Comp_m = ();
    my $size_m = 0;
    if ($NP_m && ($ToksNP_m >= $LongNPsize_m)) {
	$Comp_m = $NP_m;
	$size_m = $ToksNP_m;
    } else {
	$Comp_m = $FullNP_m;
	$size_m = $LongNPsize_m;
    }

    # Longest proper noun of the second entity
    if ($FullNPEnt{$Mentions{$i}}) {
	my %size = ();
	foreach my $f(@{$FullNPEnt{$Mentions{$i}}}) {
	    if ($f =~ /_/) {
		my (@size) = split("_", $f);
		$size{$f} = $#size + 1;
		my $first = 0;
		foreach my $n(sort {$size{$b} <=> $size{$a}} keys %size ) {
		    if (!$first) {
			$LongNP_i = $n;
			$LongNPsize_i = $size{$n};
		    }
		    $first = 1;
		}
	    }
	}
	if ($LongNP_i) {
	    $FullNP_i = $LongNP_i;
	} else {
	    $LongNPsize_i = 1;
	}
    }

    # Entity to compare
    # (from entity or mention)
    my $Comp_i = ();
    my $size_i = 0;

    if ($NP_i && ($ToksNP_i >= $LongNPsize_i)) {
	$Comp_i = $NP_i;
	$size_i = $ToksNP_i;
    } else {
	$Comp_i = $FullNP_i;
	$size_i = $LongNPsize_i;
    }

    # Verifies if the longest proper noun of an entity
    # contains all the tokens (except stop words) of the
    # longest proper noun of the other entity

    # Might have same start (not ending)
    # "Guille - Guillermo" vs "Mazinho - Zinho"
    # Not in PrepPhrases: "Portugal: LOC - Campeonato_de_Portugal: MISC"

    #if ($size_m != $size_i) {
	if ( ($Comp_m =~ /_(de|por|em|en|do|da|dos|das)_?(o|os|a|as|el|la|los|las|the)?_$Comp_i/) ||
	     ($Comp_i =~ /_(de|por|em|en|do|da|dos|das)_?(o|os|a|as|el|la|los|las|the)?_$Comp_m/) && $size_m != $size_i) {
	    $CompNP = 0;
	}
	elsif ($Comp_m =~ /(^|_|-)$Comp_i/ || $Comp_i =~ /(^|_|-)$Comp_m/ && $size_m != $size_i) {
	    unless ( ($Comp_m =~ /(^|_)jr./ && $Comp_i !~ /(^|_)jr./) ||
		     ($Comp_m !~ /(^|_)jr./ && $Comp_i =~ /(^|_)jr./) ){
		$CompNP = 1;
	    }
	}

#(previous)
#    if ($Comp_m =~ /(^|_|-)$Comp_i/ || $Comp_i =~ /(^|_|-)$Comp_m/) {
#	unless ( ($Comp_m =~ /(^|_)jr./ && $Comp_i !~ /(^|_)jr./) ||
#		 ($Comp_m !~ /(^|_)jr./ && $Comp_i =~ /(^|_)jr./) ){
#	    $CompNP = 1;
#	}
#    }
    #}
	# Token comparison (same order)
	else {
	    my $Comp_a = ();
	    my $Comp_b = ();
	    my $size_a = ();
	    my $size_b = ();
	    if ($size_m < $size_i) {
		$Comp_a = $Comp_m;
		$Comp_b = $Comp_i;
		$size_a = $size_m;
		$size_b = $size_i;
	    } elsif ($size_i < $size_m) {
		$Comp_a = $Comp_i;
		$Comp_b = $Comp_m;
		$size_a = $size_i;
		$size_b = $size_m;
	    }
	    
	    my (@Toks_m) = split("_", $Comp_a);
	    my (@Toks_i) = split("_", $Comp_b);
	    
	    my $found = 0;
	    my $all = 0;
	    my $ci = 0;
	    foreach my $tm(@Toks_m) {
		if ($tm =~ /^(dr\.|sr\.|dra\.|sra\.|dom|dona)$/) {
		    $size_a--;
		}
		foreach my $ti(@Toks_i) {
		    if ($tm =~ /^(dr\.|sr\.|dra\.|sra\.|dom|dona)$/) {
			$all--;
		    }
		    $ci++;
		    if (($tm eq $ti) && ($ci > $found)) {
			$found = $ci;
			$all++;
			#last;
		    }
		}
	    }
	    if ($all == $size_a && $all > 0) {
		$CompNP_toks = 1;
	    }
#	}
    }
    return($CompNP, $CompNP_toks);
}

# CompareTags (NE)
##################
sub CompareTags {

    my ($m, $i) = @_;
    my $TagOK = 0;

    if ( ($Tag{$m} eq "NP00G00" && $Tag{$i} eq "NP00O00") ||
	 ($Tag{$m} eq "NP00O00" && $Tag{$i} eq "NP00G00") ) {
	$TagOK = 0;
    } else {
	$TagOK = 1;
    }
    return $TagOK;
}

# SelectTag (NE)
################
sub SelectTag {

    my ($m, $i) = @_;
    my %Count = ();
    my $Tag_m = ();
    my $Tag_i = ();
    my $count_m = 0;
    my $count_i = 0;
    my $final_tag = ();

    # Get tag from mention/entity
    # First (m)
    if ($Mentions{$m} eq "X") {
	$Tag_m = $Tag{$m};
	$count_m = 1;
    } else {
	foreach my $e(@{$Mentions{$m}}) {
	    # Get the most frequent tag of the entity
	    if ($Tag{$e} eq "NP00SP0") {
		$Count{$Tag{$e}}++;
	    } elsif ($Tag{$e} eq "NP00O00") {
		$Count{$Tag{$e}}++;
	    } elsif ($Tag{$e} eq "NP00G00") {
		$Count{$Tag{$e}}++;
	    } elsif ($Tag{$e} eq "NP00V00") {
		$Count{$Tag{$e}}++;
	    }
	}
	my $first = 0;
	foreach my $n(sort {$Count{$b} <=> $Count{$a}} keys %Count ) {
	    if (!$first) {
		$Tag_m = $n;
		$count_m = $Count{$n};
	    }
	    $first = 1;
	}
    }

    # Second (i)
    if ($Mentions{$i} eq "X") {
	$Tag_i = $Tag{$i};
	$count_i = 1;
    } else {
	foreach my $e (@{$Mentions{$i}}) {
	    # Get the most frequent tag of the entity	    
	    if ($Tag{$e} eq "NP00SP0") {
		$Count{$Tag{$e}}++;
	    } elsif ($Tag{$e} eq "NP00O00") {
		$Count{$Tag{$e}}++;
	    } elsif ($Tag{$e} eq "NP00G00") {
		$Count{$Tag{$e}}++;
	    } elsif ($Tag{$e} eq "NP00V00") {
		$Count{$Tag{$e}}++;
	    }
	}
	my $first = 0;
	foreach my $n(sort {$Count{$b} <=> $Count{$a}} keys %Count ) {
	    if (!$first) {
		$Tag_i = $n;
		$count_i = $Count{$n};
	    }
	    $first = 1;
	}
    }

    # Select most frequent
    # Tag of the longest entity: NER is more reliable
    # with larger entities: "Brandia: ORG vs Manuel_Brandia: PER"
    # (1º: tag of entity_i; 2º: tag of mention_i)

    #
    if ($Mentions{$m} eq "X" && $MentionsForm{$m} =~ /_/ && $MentionsForm{$i} !~ /_/) {
	$final_tag = $Tag_m;
    } elsif ($Mentions{$i} eq "X" && $MentionsForm{$i} =~ /_/ && $MentionsForm{$m} !~ /_/) {
	$final_tag = $Tag_i;
    }
    #
    elsif ($count_m > $count_i) {
	$final_tag = $Tag_m;
    } elsif ($count_i > $count_m) {
	$final_tag = $Tag_i;
    } else {
	if ($TagEnt{$Mentions{$i}}) {
	    $final_tag = $TagEnt{$Mentions{$i}};
	} elsif ($TagEnt{$Mentions{$m}}) {
	    $final_tag = $TagEnt{$Mentions{$m}};
	} else {
	    if (length($MentionsForm{$m}) > length($MentionsForm{$i})) {
		$final_tag = $Tag_m;
	    } elsif (length($MentionsForm{$i}) > length($MentionsForm{$m})) {
		$final_tag = $Tag_i;
	    }
	    else {
		$final_tag = $Tag_i; #default
	    }
	}
    }
    return $final_tag;
}

1;
